import sys, os
from ams_instrument import ams_instrument
import logging
import nidcpower
from ams_smu import ams_SMU_instrument

my_logger = logging.getLogger(__name__)


class ams_PXI_SMU(ams_SMU_instrument):
    """
    This class implements the SMU functionality for PXI SMU device
    """
    def __init__(self, device_dict: dict) -> object:
        """
        This method initializes the SMU based on the HW configuration file
        Device dict example:
        {
            "Instrument_name": "PXI_SMU_6V_500mA_slot02",
            "Type": "PXIe-4143",
            "Details": {
                "Voltage_max": 6,
                "Current_max": 500,
                "Reset": 1,
                "Channels": [
                    "0",
                    "1"
                ] 
            }
        }
        Args:
            device_dict (dict): dictionary of instrument settings

        Returns:
            object: ams_SMU_instrument object
        """
        self.ni_device_dict = device_dict
        ni_max_alias = device_dict['Instrument_name']
        ni_channels = device_dict['Details']['Channels']
        ni_reset = device_dict['Details']['Reset'] == 1
        try:
            self.sense = device_dict['Details']['sense']
        except KeyError:
            self.sense = "local"
        self.smu_handles = {}
        for channel in ni_channels:
            self.smu_handles[channel] = nidcpower.Session(ni_max_alias, channels = channel, reset = ni_reset)
            self.smu_handles[channel].source_mode = nidcpower.SourceMode.SINGLE_POINT
            self.smu_handles[channel].measure_when = nidcpower.MeasureWhen.ON_DEMAND
            self.smu_handles[channel].output_function = nidcpower.OutputFunction.DC_VOLTAGE
            # self.smu_handles[channel].current_limit = 10e-3
            self.smu_handles[channel].current_level_autorange = False
            self.smu_handles[channel].voltage_level_autorange = False
            self.smu_handles[channel].current_limit_autorange = False
            self.smu_handles[channel].voltage_limit_autorange = False
            self.smu_handles[channel].source_delay = 0.1
            self.smu_handles[channel].current_level_range = 10e-3
            self.smu_handles[channel].voltage_level = 0
            self.smu_handles[channel].output_enabled = False
            if self.sense == "remote":
                self.smu_handles[channel].sense = nidcpower.Sense.REMOTE
            self.smu_handles[channel].initiate()


    def configure_ranges(
            self,
            channel_name: str,
            voltage_range: float=None,
            current_range: float=None,
            voltage_limit: float=None,
            current_limit: float=None
        ):
        self.smu_handles[channel_name].abort()
        if voltage_range is not None:
            self.smu_handles[channel_name].voltage_level_range = voltage_range
            self.smu_handles[channel_name].voltage_limit_range = voltage_range
        if current_range is not None:
            self.smu_handles[channel_name].current_level_range = current_range
            self.smu_handles[channel_name].current_limit_range = current_range
        if voltage_limit is not None:
            self.smu_handles[channel_name].voltage_limit = voltage_limit
        if current_limit is not None:
            self.smu_handles[channel_name].current_limit = current_limit
        self.smu_handles[channel_name].commit()
        self.smu_handles[channel_name].initiate()
        """configure_ranges 
        This method configures ranges and limits for the smu channel

        Args:
            channel_name (str): channel name to be configured
            voltage_range (float, optional): voltage range - if not given, no new limit is set. Defaults to None.
            current_range (float, optional): current range - if not given, no new limit is set. Defaults to None.
            voltage_limit (float, optional): voltage limit - if not given, no new limit is set. Defaults to None.
            current_limit (float, optional): current limit - if not given, no new limit is set. Defaults to None.
        """


    def source_voltage(
        self, 
        voltage_level: float, 
        current_limit: float, 
        channel_name: str,
        voltage_range: float=None,
        current_range: float=None
    ):
        """source_voltage 
        This method sources out voltage to a certain channel with a defined current limit

        Args:
            voltage_level (float): SMU terminal voltage
            current_limit (float): SMU current limit
            channel_name (str): Name of the channel on SMU in use
            voltage_range (float, optional): voltage range - if not given, no new limit is set. Defaults to None.
            current_range (float, optional): current range - if not given, no new limit is set. Defaults to None.
        """
        
        act_func = nidcpower.OutputFunction.DC_VOLTAGE
        self._check_ranges_and_func(channel_name, act_func, voltage_range, current_range)
        self.smu_handles[channel_name].current_limit = current_limit
        self.smu_handles[channel_name].voltage_level = voltage_level


    def _check_ranges_and_func(
        self,
        channel_name: str,
        function: nidcpower.OutputFunction,
        voltage_range: float,
        current_range: float
        ):
        """
        This method checks if the ranges and the output function of a given channel is set correctly and if not changes it.

        Args:
            channel_name (str): name of the channel
            function (nidcpower.OutputFunction): desired output function
            voltage_range (float): voltage range
            current_range (float): current range
        """
        func_ok = self.smu_handles[channel_name].output_function != function
        vrange_ok = self.smu_handles[channel_name].voltage_level_range != voltage_range
        crange_ok = self.smu_handles[channel_name].current_level_range != current_range
        self.smu_handles[channel_name].abort()
        if func_ok or vrange_ok or crange_ok:
            self.smu_handles[channel_name].output_function = function  
        if voltage_range is not None:
            my_logger.debug("setting voltage range")
            # print("setting voltage range")
            self.smu_handles[channel_name].voltage_limit_range = voltage_range
            self.smu_handles[channel_name].voltage_level_range = voltage_range
        if current_range is not None:
            my_logger.debug("setting current range")
            # print("setting current range")
            self.smu_handles[channel_name].current_limit_range = current_range
            self.smu_handles[channel_name].current_level_range = current_range
        self.smu_handles[channel_name].commit()
        self.smu_handles[channel_name].initiate()



    def source_current(
        self, 
        current_level: float, 
        voltage_limit: float, 
        channel_name: str,
        voltage_range: float=None,
        current_range: float=None
    ):
        """source_current 
        This method sources out current to a certain channel with a defined voltage limit

        Args:
            current_level (float): SMU terminal current
            voltage_limit (float): SMU voltage limit
            channel_name (str): Name of the channel on SMU in use
            voltage_range (float, optional): voltage range - if not given, no new limit is set. Defaults to None.
            current_range (float, optional): current range - if not given, no new limit is set. Defaults to None.
        """
        
        act_func = nidcpower.OutputFunction.DC_CURRENT
        self._check_ranges_and_func(channel_name, act_func, voltage_range, current_range)
        self.smu_handles[channel_name].current_level = current_level
        self.smu_handles[channel_name].voltage_limit = voltage_limit

    @classmethod
    def _convert_parameters(
        self,
        duration: float=None, 
        sampling_rate: float=None, 
        number_of_samples: float=None
    ) -> list:
        """_convert_parameters 
        This method calculates necessary parameters of multipoint measurement in case of different scenarios of input parameter combintaions

        Args:
            duration (float, optional): measurement duration in seconds. Defaults to None.
            sampling_rate (float, optional): sampling rate in sample/seconds. Defaults to None.
            number_of_samples (float, optional): number of points to measure. Defaults to None.

        Returns:
            list: list of calculated dureation, sampling_rate and number_of_samples
        """
        my_logger.debug(f'duration: {duration}, sampling rate: {sampling_rate}, number of samples:{number_of_samples}')
        if duration == None and sampling_rate != None and number_of_samples != None:
            my_logger.debug('Duration is not given')
            my_sampling_rate = sampling_rate
            my_duration = int(sampling_rate * number_of_samples)
            my_number_of_samples = number_of_samples
        elif duration != None and sampling_rate == None and number_of_samples != None:
            my_logger.debug('Sampling rate is not given')
            my_sampling_rate = int(duration / number_of_samples)
            my_duration = duration
            my_number_of_samples = number_of_samples
        elif duration != None and sampling_rate != None and number_of_samples == None:
            my_logger.debug('Number of samples is not given')
            my_sampling_rate = sampling_rate
            my_duration = duration
            my_number_of_samples = int(duration / sampling_rate)
        my_duration = 1
        return my_duration, my_sampling_rate, my_number_of_samples


    def measure_voltage(
        self, 
        channel_name: str,
        range: str=None, 
    ) -> float:
        """measure_voltage 
        This method measures the voltage on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.

        Returns:
            float: measured voltage
        """

        my_logger.debug('Single point measurement')
        self.smu_handles[channel_name].abort()
        self.smu_handles[channel_name].measure_when = nidcpower.MeasureWhen.ON_DEMAND
        self.smu_handles[channel_name].commit()
        self.smu_handles[channel_name].initiate()
        return self.smu_handles[channel_name].measure(nidcpower.MeasurementTypes.VOLTAGE)

    def measure_voltage_cont(
        self, 
        channel_name: str,
        range: str=None, 
        duration: float=None, 
        sampling_rate: float=None, 
        number_of_samples: float=None
    ) -> list:
        """measure_voltage 
        This method measures the voltage on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.
            duration (float, optional): duration of measurement in seconds. Defaults to None.
            sampling_rate (float, optional): sampling rate. Defaults to None.
            number_of_samples (float, optional): number of samples to be measured. Defaults to None.

        Returns:
            list: list of measured voltages
        """
        my_logger.debug('multi-point measurement')
        my_duration, my_sampling_rate, my_number_of_samples = self._convert_parameters(duration, sampling_rate, number_of_samples)
        my_logger.debug(f'{my_duration} - {my_sampling_rate} - {my_number_of_samples}')
        self.smu_handles[channel_name].abort()
        self.smu_handles[channel_name].measure_record_length = my_number_of_samples
        self.smu_handles[channel_name].measure_record_length_is_finite = True
        self.smu_handles[channel_name].measure_when = nidcpower.MeasureWhen.ON_MEASURE_TRIGGER
        self.smu_handles[channel_name].measure_trigger_type = nidcpower.TriggerType.SOFTWARE_EDGE
        self.smu_handles[channel_name].commit()
        self.smu_handles[channel_name].initiate()
        self.smu_handles[channel_name].send_software_edge_trigger(nidcpower.SendSoftwareEdgeTriggerType.START)
        measurements = self.smu_handles[channel_name].fetch_multiple(my_number_of_samples, my_duration + 1)
        voltages = []
        for measurement in measurements:
            voltages.append(measurement.voltage)
        return voltages


    def measure_current(
        self, 
        channel_name: str,
        range: str=None, 
    ) -> float:
        """
        This method measures the current on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.

        Returns:
            float: measured current
        """

        self.smu_handles[channel_name].abort()
        self.smu_handles[channel_name].measure_when = nidcpower.MeasureWhen.ON_DEMAND
        self.smu_handles[channel_name].commit()
        self.smu_handles[channel_name].initiate()
        return self.smu_handles[channel_name].measure(nidcpower.MeasurementTypes.CURRENT)


    def measure_current_cont(
        self, 
        channel_name: str,
        range: str=None, 
        duration: float=None, 
        sampling_rate: float=None, 
        number_of_samples: float=None
    ) -> list:
        """
        This method measures the current on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.
            duration (float, optional): duration of measurement in seconds. Defaults to None.
            sampling_rate (float, optional): sampling rate. Defaults to None.
            number_of_samples (float, optional): number of samples to be measured. Defaults to None.

        Returns:
            list: list of measured currents
        """
        my_duration, my_sampling_rate, my_number_of_samples = self._convert_parameters(duration, sampling_rate, number_of_samples)
        my_logger.debug(f'{my_duration} - {my_sampling_rate} - {my_number_of_samples}')
        self.smu_handles[channel_name].abort()
        self.smu_handles[channel_name].measure_record_length = my_number_of_samples
        self.smu_handles[channel_name].measure_record_length_is_finite = True
        self.smu_handles[channel_name].measure_when = nidcpower.MeasureWhen.ON_MEASURE_TRIGGER
        self.smu_handles[channel_name].measure_trigger_type = nidcpower.TriggerType.SOFTWARE_EDGE
        self.smu_handles[channel_name].commit()
        self.smu_handles[channel_name].initiate()
        self.smu_handles[channel_name].send_software_edge_trigger(nidcpower.SendSoftwareEdgeTriggerType.START)
        measurements = self.smu_handles[channel_name].fetch_multiple(my_number_of_samples, my_duration + 1)
        currents = []
        for measurement in measurements:
            currents.append(measurement.current)
        return currents


    def output_enable(self, channel_name: str, enabled: bool=False):
        """output_enable 
        This function changes the state of the device output

        Args:
            channel_name (str): channel name
            enabled (bool, optional): enable state. Defaults to False.
        """
        self.smu_handles[channel_name].output_enabled = enabled


    def __del__(self):
        
        for channel in self.ni_device_dict['Details']['Channels']:
            my_logger.debug(f'Channel to be closed: {channel}')
            self.smu_handles[channel].channels[channel].output_enabled = False
            self.smu_handles[channel].abort()
            self.smu_handles[channel].close()

    def GetInstrumentError(self) -> str:
        """IdentifyInstrument 
        Gets the identifier of the instrument as string

        Returns:
            ID_string [String]: The identification string from the instrument
        """
        return self.smu_handle._get_error()


    def IdentifyInstrument(self) -> str:
        """GetInstrumentError
        This function returns the error string from the instrument

        Returns:
            Error_string [String]: The error string from the instrument
        """
        return self.smu_handle.logical_name


    def ResetInstrumehnt(self) -> str:
        """ResetInstrumehnt 
        This function resets the instrument
        """
        return self.smu_handle.reset()