from abc import ABC, abstractmethod
import sys, os
from ams_instrument import ams_instrument
import logging
import nidcpower

my_logger = logging.getLogger(__name__)


class ams_SMU_instrument(ams_instrument.ams_instrument):
    """
    This interface defines the functionality of an SMU instrument
    """
    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        Initializes the ams_SMU_instrument object

        Args:
            device_dict (dict): dictionary of instrument settings

        Returns:
            object: ams_SMU_instrument object
        """
        pass


    @abstractmethod
    def configure_ranges(
        self,
        channel_name: str,
        voltage_range: float=None,
        current_range: float=None,
        voltage_limit: float=None,
        current_limit: float=None
    ):
        """configure_ranges 
        This method configures ranges and limits for the smu channel

        Args:
            channel_name (str): channel name to be configured
            voltage_range (float, optional): voltage range - if not given, no new limit is set. Defaults to None.
            current_range (float, optional): current range - if not given, no new limit is set. Defaults to None.
            voltage_limit (float, optional): voltage limit - if not given, no new limit is set. Defaults to None.
            current_limit (float, optional): current limit - if not given, no new limit is set. Defaults to None.
        """


    @abstractmethod
    def source_voltage(
        self, 
        voltage_level: float, 
        current_limit: float, 
        channel_name: str,
        voltage_range: float=None,
        current_range: float=None
    ):
        """source_voltage 
        This method sources out voltage to a certain channel with a defined current limit

        Args:
            voltage_level (float): SMU terminal voltage
            current_limit (float): SMU current limit
            channel_name (str): Name of the channel on SMU in use
            voltage_range (float, optional): voltage range - if not given, no new limit is set. Defaults to None.
            current_range (float, optional): current range - if not given, no new limit is set. Defaults to None.
        """
        pass


    @abstractmethod
    def source_current(
        self, 
        current_level: float, 
        voltage_limit: float, 
        channel_name: str,
        voltage_range: float=None,
        current_range: float=None
    ):
        """source_current 
        This method sources out current to a certain channel with a defined voltage limit

        Args:
            current_level (float): SMU terminal current
            voltage_limit (float): SMU voltage limit
            channel_name (str): Name of the channel on SMU in use
            voltage_range (float, optional): voltage range - if not given, no new limit is set. Defaults to None.
            current_range (float, optional): current range - if not given, no new limit is set. Defaults to None.
        """
        pass


    @abstractmethod
    def measure_voltage(
        self, 
        channel_name: str,
        range: str=None, 
    ) -> float:
        """measure_voltage 
        This method measures the voltage on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.

        Returns:
            float: measured voltage
        """
        pass


    @abstractmethod
    def measure_voltage_cont(
        self, 
        channel_name: str,
        range: str=None, 
        duration: float=None, 
        sampling_rate: float=None, 
        number_of_samples: float=None
    ) -> list:
        """measure_voltage 
        This method measures the voltage on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.
            duration (float, optional): duration of measurement in seconds. Defaults to None.
            sampling_rate (float, optional): sampling rate. Defaults to None.
            number_of_samples (float, optional): number of samples to be measured. Defaults to None.

        Returns:
            list: list of measured voltages
        """
        pass


    @abstractmethod
    def measure_current(
        self, 
        channel_name: str,
        range: str=None, 
    ) -> float:
        """
        This method measures the current on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.

        Returns:
            float: measured current
        """
        pass


    @abstractmethod
    def measure_current_cont(
        self, 
        channel_name: str,
        range: str=None, 
        duration: float=None, 
        sampling_rate: float=None, 
        number_of_samples: float=None
    ) -> list:
        """
        This method measures the current on a certain channel with given parameters

        Args:
            channel_name (str): SMU channel name
            range (str, optional): measurement range. Defaults to None.
            duration (float, optional): duration of measurement in seconds. Defaults to None.
            sampling_rate (float, optional): sampling rate. Defaults to None.
            number_of_samples (float, optional): number of samples to be measured. Defaults to None.

        Returns:
            list: list of measured currents
        """
        pass


    @abstractmethod
    def output_enable(self, channel_name: str, enabled: bool=False):
        """output_enable 
        This function changes the state of the device output

        Args:
            channel_name (str): channel name
            enabled (bool, optional): enable state. Defaults to False.
        """
        pass


    @abstractmethod
    def GetInstrumentError(self) -> str:
        """IdentifyInstrument 
        Gets the identifier of the instrument as string

        Returns:
            ID_string [String]: The identification string from the instrument
        """
        pass


    @abstractmethod
    def IdentifyInstrument(self) -> str:
        """GetInstrumentError
        This function returns the error string from the instrument

        Returns:
            Error_string [String]: The error string from the instrument
        """
        pass


    @abstractmethod
    def ResetInstrumehnt(self) -> None:
        """ResetInstrumehnt 
        This function resets the instrument
        """
        pass



